#!/usr/bin/env python

"""
Moin wiki format parser.

Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.tree import Block, Break, DefItem, DefTerm, FontStyle, Heading, \
                            Larger, ListItem, Monospace, Region, Rule, Smaller, \
                            Subscript, Superscript, TableAttr, TableAttrs, \
                            TableCell, TableRow, Text, Underline
import re

# Regular expressions.

syntax = {
    # Page regions:
    "regionstart"   : r"((^\s*)([{]{3,}))",                         # {{{...
    "regionend"     : r"^\s*([}]{3,})",                             # }}}...
    "header"        : r"#!(.*?)\n",                                 # #! char-excl-nl

    # Region contents:
    # Line-oriented patterns:
                      # blank line
    "break"         : r"^(\s*?)\n",
                      # ws... expecting text ::
    "defterm"       : r"^(\s+)(?=.+?::)",
                      # ws... expecting :: ws...
    "defterm_empty" : r"^(\s+)(?=::\s+)",
                      # [ws...] =... ws... expecting headingend
    "heading"       : r"^(\s*)(?P<x>=+)(\s+)(?=.*?\s+(?P=x)\s*\n)",
                      # ws... list-item [ws...]
    "listitem"      : r"^(\s+)(\*)(\s*)",
                      # ws... number-item ws...
    "listitem_num"  : r"^(\s+)(\d+\.)(\s+)",
                      # ws... alpha-item ws...
    "listitem_alpha": r"^(\s+)([aA]\.)(\s+)",
                      # ws... roman-item ws...
    "listitem_roman": r"^(\s+)([iI]\.)(\s+)",
                      # ws... dot-item [ws...]
    "listitem_dot"  : r"^(\s+)(\.)(\s*)",
                      # ||
    "tablerow"      : r"^\|\|",

    # Region contents:
    # Inline patterns:
    "fontstyle"     : r"('{2,6})",
    "larger"        : r"~\+",
    "monospace"     : r"`",
    "rule"          : r"(-----*)",                                  # ----...
    "smaller"       : r"~-",
    "sub"           : r",,",
    "super"         : r"\^",
    "underline"     : r"__",

    # Inline contents:
    "largerend"     : r"\+~",
    "monospaceend"  : r"`",
    "smallerend"    : r"-~",
    "subend"        : r",,",
    "superend"      : r"\^",
    "underlineend"  : r"__",

    # Heading contents:
    "headingend"    : r"(\s+)(=+)(\s*\n)",                          # ws... =... [ws...] nl

    # List contents:
    "deftermend"    : r"::(\s*?\n)",
    "deftermsep"    : r"::(\s+)",
    "listitemend"   : r"^",                                         # next line

    # Table contents:
    "tableattrs"    : r"<",
    "tablecell"     : r"\|\|",
    "tableend"      : r"(\s*?)^",                                   # [ws...] next line

    # Table attributes:
    "tableattrsend" : r">",
    "halign"        : r"([(:)])",
    "valign"        : r"([v^])",
    "colour"        : r"(\#[0-9A-F]{6})",
    "colspan"       : r"-(\d+)",
    "rowspan"       : r"\|(\d+)",
    "width"         : r"(\d+%)",
    "attrname"      : r"((?![-\d])[-\w]+)",                         # not-dash-or-digit dash-or-word-char...
    "attrvalue"     : r"""=(?P<x>['"])(.*?)(?P=x)""",
    }

# Define pattern details.

table_pattern_names = ["attrname", "colour", "colspan", "halign", "rowspan", "tableattrsend", "valign", "width"]

inline_pattern_names = ["fontstyle", "larger", "monospace", "smaller", "sub", "super", "underline"]

def inline_patterns_for(name):
    names = inline_pattern_names[:]
    names[names.index(name)] = "%send" % name
    return names

# Define patterns for the regular expressions.

patterns = {}
for name, value in syntax.items():
    patterns[name] = re.compile(value, re.UNICODE | re.MULTILINE)



# Tokenising functions.

class TokenStream:

    "A stream of tokens taken from a string."

    def __init__(self, s):
        self.s = s
        self.pos = 0
        self.match = None
        self.matching = None

    def rewind(self, length):

        "Rewind in the string by 'length'."

        self.pos -= min(length, self.pos)

    def read_until(self, pattern_names, remaining=True):

        """
        Find the first match for the given 'pattern_names'. Return the text
        preceding any match, the remaining text if no match was found, or None
        if no match was found and 'remaining' is given as a false value.
        """

        first = None
        self.matching = None

        # Find the first matching pattern.

        for pattern_name in pattern_names:
            match = patterns[pattern_name].search(self.s, self.pos)
            if match:
                start, end = match.span()
                if self.matching is None or start < first:
                    first = start
                    self.matching = pattern_name
                    self.match = match

        if self.matching is None:
            if remaining:
                return self.s[self.pos:]
            else:
                return None
        else:
            return self.s[self.pos:first]

    def read_match(self, group=1):

        """
        Return the matched text, updating the position in the stream. If 'group'
        is specified, the indicated group in a match will be returned.
        Typically, group 1 should contain all pertinent data, but groups defined
        within group 1 can provide sections of the data.
        """

        if self.match:
            _start, self.pos = self.match.span()
            try:
                return self.match.group(group)
            except IndexError:
                return ""
        else:
            self.pos = len(self.s)
            return None



# Parser functions.

def parse_page(s):

    """
    Parse page text 's'. Pages consist of regions delimited by markers.
    """

    return parse_region(TokenStream(s))

def parse_region(items, level=0, indent=0):

    """
    Parse the data provided by 'items' to populate a region with the given
    'level' at the given 'indent'.
    """

    region = Region([], level, indent)

    # Parse section headers.

    parse_region_header(items, region)

    # Parse section body.

    if region.is_transparent():
        parse_region_wiki(items, region)
    else:
        parse_region_opaque(items, region)

    return region

def parse_region_header(items, region):

    """
    Parse the region header from the 'items', setting it for the given 'region'.
    """

    if items.read_until(["header"], False) == "": # None means no header
        region.type = items.read_match()

def parse_region_wiki(items, region):

    "Parse the data provided by 'items' to populate a wiki 'region'."

    new_block(region)
    parse_region_details(items, region, inline_pattern_names + [
        "break", "heading",
        "defterm", "defterm_empty",
        "listitem", "listitem_alpha", "listitem_dot", "listitem_num",
        "listitem_roman",
        "regionstart", "regionend",
        "rule",
        "tablerow",
        ])

def parse_region_opaque(items, region):

    "Parse the data provided by 'items' to populate an opaque 'region'."

    parse_region_details(items, region, ["regionend"])

def parse_region_details(items, region, pattern_names):

    "Parse 'items' within 'region' searching using 'pattern_names'."

    try:
        while True:

            # Obtain text before any marker or the end of the input.

            preceding = items.read_until(pattern_names)
            if preceding:
                region.append_inline(Text(preceding))

            # End of input.

            if not items.matching:
                break

            # Obtain any feature.

            feature = items.read_match()
            handler = handlers.get(items.matching)

            # Handle each feature or add text to the region.

            if handler:
                handler(items, region)
            else:
                region.append_inline(Text(feature))

    except StopIteration:
        pass

    region.normalise()

def end_region(items, region):

    "End the parsing of 'region'."

    raise StopIteration

def parse_attrname(items, attrs):

    "Handle an attribute name within 'attrs'."

    name = items.read_match()
    attr = TableAttr(name)

    preceding = items.read_until(["attrvalue"], False)
    if preceding == "":
        attr.quote = items.read_match(1)
        attr.value = items.read_match(2)

    attrs.append(attr)

def parse_break(items, region):

    "Handle a paragraph break within 'region'."

    region.add(Break())
    new_block(region)

def parse_defitem(items, region, extra=""):

    "Handle a definition item within 'region'."

    pad = items.read_match(1)
    item = DefItem([], pad, extra)
    parse_region_details(items, item, ["listitemend"])
    region.add(item)
    new_block(region)

def parse_defterm(items, region):

    "Handle a definition term within 'region'."

    pad = items.read_match(1)
    term = DefTerm([], pad)
    parse_region_details(items, term, ["deftermend", "deftermsep"])
    region.add(term)
    if items.matching == "deftermsep":
        parse_defitem(items, region)

def parse_defterm_empty(items, region):

    "Handle an empty definition term within 'region'."

    extra = items.read_match(1)
    parse_region_details(items, region, ["deftermsep"])
    parse_defitem(items, region, extra)

def parse_fontstyle(items, region):

    "Handle emphasis and strong styles."

    n = len(items.read_match(1))

    # Handle endings.

    if isinstance(region, FontStyle):
        emphasis = n in (2, 4, 5)
        strong = n in (3, 5, 6)
        active = True

        if region.emphasis and emphasis:
            active = region.close_emphasis()
            n -= 2
        if region.strong and strong:
            active = region.close_strong()
            n -= 3

        if not active:
            if n:
                items.rewind(n)
            raise StopIteration

        elif not n:
            return

    # Handle new styles.

    emphasis = n in (2, 4, 5)
    strong = n in (3, 5, 6)
    double = n in (4, 6)

    span = FontStyle([], emphasis, strong)
    if not double:
        parse_region_details(items, span, inline_pattern_names)
    region.append_inline(span)

def parse_halign(items, attrs):

    "Handle horizontal alignment within 'attrs'."

    value = items.read_match()
    attr = TableAttr("halign", value == "(" and "left" or value == ")" and "right" or "center", True)
    attrs.append(attr)

def parse_heading(items, region):

    "Handle a heading."

    start_extra = items.read_match(1)
    level = len(items.read_match(2))
    start_pad = items.read_match(3)
    heading = Heading([], level, start_extra, start_pad)
    parse_region_details(items, heading, ["headingend"] + inline_pattern_names)
    region.add(heading)
    new_block(region)

def parse_heading_end(items, heading):

    "Handle the end of a heading."

    level = len(items.read_match(2))
    if heading.level == level:
        heading.end_pad = items.read_match(1)
        heading.end_extra = items.read_match(3)
        raise StopIteration

def parse_listitem(items, region):

    "Handle a list item marker within 'region'."

    indent = len(items.read_match(1))
    marker = items.read_match(2)
    space = items.read_match(3)
    item = ListItem([], indent, marker, space)
    parse_region_details(items, item, ["listitemend"])
    region.add(item)
    new_block(region)

def parse_rule(items, region):

    "Handle a horizontal rule within 'region'."

    length = len(items.read_match(1))
    rule = Rule(length)
    region.add(rule)
    new_block(region)

def parse_section(items, region):

    "Handle the start of a new section within 'region'."

    # Parse the section and start a new block after the section.

    indent = len(items.read_match(2))
    level = len(items.read_match(3))
    region.add(parse_region(items, level, indent))
    new_block(region)

def parse_section_end(items, region):

    "Handle the end of a new section within 'region'."

    feature = items.read_match()
    if region.have_end(feature):
        raise StopIteration
    else:
        region.append_inline(Text(feature))

def parse_table_attrs(items, cell):

    "Handle the start of table attributes within 'cell'."

    attrs = TableAttrs([])
    parse_region_details(items, attrs, table_pattern_names)
    cell.attrs = attrs

def parse_table_row(items, region):

    "Handle the start of a table row within 'region'."

    row = TableRow([])

    while True:
        cell = TableCell([])
        parse_region_details(items, cell, ["tableattrs", "tablecell", "tableend"])

        # Handle the end of the row.

        if items.matching == "tableend":
            trailing = items.read_match()

            # If the cell was started but not finished, convert the row into text.

            if not row.nodes or not cell.empty():
                region.append_inline(Text("||"))

                # Convert all cells.

                for node in row.nodes:
                    region.append_inline_many(node.nodes)
                    region.append_inline(Text("||"))

                region.append_inline_many(cell.nodes)
                region.append_inline(Text(trailing))

                new_block(region)
                return

            # Append the final cell, if not empty.

            else:
                row.trailing = trailing

                if not cell.empty():
                    row.append(cell)
                break

        # A cell separator has been found.

        row.append(cell)

    region.add(row)
    new_block(region)

def parse_valign(items, attrs):

    "Handle vertical alignment within 'attrs'."

    value = items.read_match()
    attr = TableAttr("valign", value == "^" and "top" or "bottom", True)
    attrs.append(attr)

# Inline formatting handlers.

def parse_inline(items, region, cls, pattern_name):

    "Handle an inline region."

    span = cls([])
    parse_region_details(items, span, inline_patterns_for(pattern_name))
    region.append_inline(span)

parse_larger = lambda items, region: parse_inline(items, region, Larger, "larger")
parse_monospace = lambda items, region: parse_inline(items, region, Monospace, "monospace")
parse_smaller = lambda items, region: parse_inline(items, region, Smaller, "smaller")
parse_sub = lambda items, region: parse_inline(items, region, Subscript, "sub")
parse_super = lambda items, region: parse_inline(items, region, Superscript, "super")
parse_underline = lambda items, region: parse_inline(items, region, Underline, "underline")

# Table attribute handlers.

def parse_table_attr(items, attrs, pattern_name):

    "Handle a table attribute."

    value = items.read_match()
    attrs.append(TableAttr(pattern_name, value, True))

parse_colour = lambda items, cell: parse_table_attr(items, cell, "colour")
parse_colspan = lambda items, cell: parse_table_attr(items, cell, "colspan")
parse_rowspan = lambda items, cell: parse_table_attr(items, cell, "rowspan")
parse_width = lambda items, cell: parse_table_attr(items, cell, "width")

# Pattern handlers.

handlers = {
    None : end_region,
    "attrname" : parse_attrname,
    "break" : parse_break,
    "colour" : parse_colour,
    "colspan" : parse_colspan,
    "defterm" : parse_defterm,
    "defterm_empty" : parse_defterm_empty,
    "deftermend" : end_region,
    "deftermsep" : end_region,
    "fontstyle" : parse_fontstyle,
    "halign" : parse_halign,
    "heading" : parse_heading,
    "headingend" : parse_heading_end,
    "larger" : parse_larger,
    "largerend" : end_region,
    "listitemend" : end_region,
    "listitem" : parse_listitem,
    "listitem_alpha" : parse_listitem,
    "listitem_dot" : parse_listitem,
    "listitem_num" : parse_listitem,
    "listitem_roman" : parse_listitem,
    "monospace" : parse_monospace,
    "monospaceend" : end_region,
    "regionstart" : parse_section,
    "regionend" : parse_section_end,
    "rowspan" : parse_rowspan,
    "rule" : parse_rule,
    "smaller" : parse_smaller,
    "smallerend" : end_region,
    "sub" : parse_sub,
    "subend" : end_region,
    "super" : parse_super,
    "superend" : end_region,
    "tableattrs" : parse_table_attrs,
    "tableattrsend" : end_region,
    "tablerow" : parse_table_row,
    "tablecell" : end_region,
    "tableend" : end_region,
    "underline" : parse_underline,
    "underlineend" : end_region,
    "valign" : parse_valign,
    "width" : parse_width,
    }

def new_block(region):

    "Start a new block in 'region'."

    block = Block([])
    region.add(block)



# Top-level functions.

parse = parse_page

# vim: tabstop=4 expandtab shiftwidth=4
